
/*
   This is the firmware to control the animatronic tentacle mechanism. It uses a custom controller using a Teensy 3.6 or 3.5
*/
#include <PID_v1.h>
#include "Serialcom.cpp"
#include <Servo.h>

SerialCommand Ser;

Servo servo1;
Servo servo2;

uint16_t servo_pos1, servo_pos2;
uint16_t servo1_range[2] = {4000, 6600}; //the range of the input for the servo (servo goes from 0-180) (4000, 7600)
uint16_t servo2_range[2] = {7500, 2600}; //the range of the input for the servo (servo goes from 0-180) (6300, 1400)
uint32_t servo1_micros[2] = {1050, 1750};
uint32_t servo2_micros[2] = {700, 2300};
uint16_t servo_history[2][3]; //used to calculate rolling average
uint16_t servo_from_pos[2];
uint16_t servo_to_pos[2];
uint32_t servo_start_time;
uint8_t servo_gripper_active = 0;
uint8_t servo_rotation_active = 0;
uint32_t servo_update_target;
uint32_t servo_update_delay = 20;


#define number_of_motors 8
//motor pins go from M0 to M7.
const uint8_t motor_step[number_of_motors] = {13, 15, 12, 11, 10, 9, 22, 23};
const uint8_t motor_dir[number_of_motors] = {32, 30, 45, 44, 43, 24, 26, 40};
const uint8_t motor_enable[number_of_motors] = {31, 46, 29, 28, 27, 25, 42, 41};
//endstops go from X- to Z+
const uint8_t motor_endstop[number_of_motors] = {33, 47, 34, 48, 35, 49, 36, 50};
const uint8_t motor_pot[number_of_motors] = {A0, A2, A3, A4, A5, A6, A7, A10};
const uint8_t activation_switch = 8;



uint8_t activation_switch_history;
uint32_t activation_delay = 10000000; //the delay between checking of the button
uint32_t activation_target;
uint8_t motion_activated = 0;

const uint32_t motor_range[number_of_motors] = {66000, 88000, 66000, 66000, 66000, 66000, 44000, 44000}; //how far the motors need to step from 0 to full (homing in the middle)
const uint32_t motor_accuracy = 100; //how close in steps the position needs to be to the target

uint32_t step_delay = 1000; //the minimum delay between each step
int32_t step_speed[number_of_motors];
int32_t motor_min_velocity[number_of_motors] = {100, 100, 100, 100, 100, 100, 67, 67};
int32_t motor_max_velocity[number_of_motors] = {3000, 3000, 3000, 3000, 3000, 3000, 2000, 2000};
uint32_t step_max_offset = 500;
uint32_t step_history[number_of_motors]; //the time of the last step
uint16_t input_target_position[number_of_motors] = {4096, 4096, 4096, 4096, 4096, 4096, 4096, 4096}; //where the motors should be
int32_t motor_target_offset[number_of_motors];
int32_t motor_target_position[number_of_motors];
int32_t motor_current_position[number_of_motors];
int8_t motor_direction[number_of_motors];
int32_t motor_current_velocity[number_of_motors];
int32_t motor_target_velocity[number_of_motors];
#define serial_buffer_size 128
uint16_t serial_update_speed = 100; //in milliseconds
char serial_input_buffer[serial_buffer_size];
uint16_t serial_read_pos;
#define serial_decode_size 16
uint8_t serialCommandEcho = 1; //used to return the value decoded from command
uint32_t serial_command;
int32_t serial_small_value[8];
int32_t serial_position[8];
char serial_decode_array[serial_decode_size];
uint8_t step_required = 0;
uint8_t reread_serial = 0;

uint32_t disable_target;
uint32_t disable_delay = 10000; //the time in millis without update before the motors are dissabled
uint8_t motor_enabled = 0;

uint32_t analog_read_target;
uint32_t analog_read_delay = 100;
uint16_t analog_read_value[number_of_motors];
uint32_t analog_center[number_of_motors] = {4096, 4096, 4096, 4096, 4096, 4096, 4096, 4096}; //<--------------------- integrate these in the calculation
uint32_t analog_sweep[number_of_motors] = {8191, 8191, 8191, 8191, 8191, 8191, 8191, 8191}; //how far the whole sweep is

double pid_input[number_of_motors];
double pid_output[number_of_motors];
double pid_output_history[number_of_motors];
double pid_setpoint[number_of_motors];
double pid_kp = -3.5;
double pid_ki = 0.0;
double pid_kd = 1.5;
double pid_max_delta = 50; //the max change per cycle

uint32_t pid_read_target;
uint32_t pid_read_delay = 10;


PID pid_motor[] = {
  PID(&pid_input[0], &pid_output[0], &pid_setpoint[0], pid_kp, pid_ki, pid_kd, DIRECT),
  PID(&pid_input[1], &pid_output[1], &pid_setpoint[1], pid_kp, pid_ki, pid_kd, DIRECT),
  PID(&pid_input[2], &pid_output[2], &pid_setpoint[2], pid_kp, pid_ki, pid_kd, DIRECT),
  PID(&pid_input[3], &pid_output[3], &pid_setpoint[3], pid_kp, pid_ki, pid_kd, DIRECT),
  PID(&pid_input[4], &pid_output[4], &pid_setpoint[4], pid_kp, pid_ki, pid_kd, DIRECT),
  PID(&pid_input[5], &pid_output[5], &pid_setpoint[5], pid_kp, pid_ki, pid_kd, DIRECT),
  PID(&pid_input[6], &pid_output[6], &pid_setpoint[6], pid_kp, pid_ki, pid_kd, DIRECT),
  PID(&pid_input[7], &pid_output[7], &pid_setpoint[7], pid_kp, pid_ki, pid_kd, DIRECT)
};



uint32_t temp_time;
uint8_t temp_val = 0;

void setup() {
  Serial.begin(115200);
  Serial.println("ATM controller");

  Ser.Begin(); //start serial connection

  analogReadResolution(13); //set resolution of analog read

  //set pinmodes for all pins
  for (uint8_t m = 0; m < number_of_motors; m++) {
    pinMode(motor_step[m], OUTPUT);
    pinMode(motor_dir[m], OUTPUT);
    pinMode(motor_enable[m], OUTPUT);
    pinMode(motor_endstop[m], INPUT);
    pinMode(motor_pot[m], INPUT);
    motor_current_position[m] = motor_range[m] / 2;

    pinMode(activation_switch, INPUT_PULLUP); //set switch to pullup

    pid_input[m] = analogRead(motor_pot[m]);
    pid_setpoint[m] = 4096.0;
    pid_motor[m].SetOutputLimits(10, 1000);
    pid_motor[m].SetSampleTime(pid_read_delay);
    pid_motor[m].SetMode(AUTOMATIC);
  }
}

void loop() {
  //serial parser
  //temporary serial parser
  if (Ser.Update() == 1) { //get serial, if 1, add new command
    /*Serial.print(Ser.GetCommand()); //send over serial debug lines
      Serial.print(" ");
      for (uint8_t s = 0; s < 8; s++) {
      Serial.print(Ser.GetSmallValue(s));
      Serial.print(",");
      }
      Serial.println("");//*/
    SerialExecute(); //get command and execute it
  }

  //check if motors need to move
  for (uint8_t m = 0; m < 8; m++) {
    //calculate target position based on input position
    motor_target_position[m] = map(input_target_position[m], 0, 8191, 0, motor_range[m]);

    //if the motor is not on target
    motor_target_offset[m] = motor_current_position[m] - motor_target_position[m];

    if (motion_activated == 3) { //if the motion is active, allow movement
      if (abs(motor_target_offset[m]) > motor_accuracy) { //if motion is required
        disable_target = millis() + disable_delay;
        //Serial.println("Reset delay");

        if (micros() > step_history[m] + abs(step_speed[m])) { //check if motor can be stepped
          step_history[m] = micros(); //set new history

          if (motor_direction[m] == -1) { //if negative step
            digitalWrite(motor_dir[m], 0);
            motor_current_position[m] --;
          }
          else { //if positive step
            digitalWrite(motor_dir[m], 1);
            motor_current_position[m] ++;
          }

          //mark a step
          step_required = 1;
          SetStep(m);
          //Serial.println("Step");
        }
      }
    }
  }

  //step motors that were marked
  if (step_required == 1) {
    step_required = 0;
    Step();
  }

  if (millis() > disable_target) {
    //Serial.println("disable motors");
    SetEnable(0);
    motor_enabled = 0;
    servo1.detach();
    servo2.detach();
  }
  else {
    if (motor_enabled == 0) {
      SetEnable(1);
      motor_enabled = 1;
      servo1.attach(35);
      servo2.attach(36);
    }
  }

  //servo update
  if (millis() > servo_update_target) {
    servo_update_target = millis() + servo_update_delay;
    //check if position is not yet reached
    if (millis() < servo_start_time + serial_update_speed) {
      //calculate new servo position
      int32_t temp_time = millis() - servo_start_time; //get elapsed time
      uint16_t temp_pos;

      //rotation calculation
      temp_pos = map(temp_time, 0, serial_update_speed, servo_from_pos[0], servo_to_pos[0]);

      //rotation
      if (servo_rotation_active == 1){
        servo1.writeMicroseconds(temp_pos);
      }

      //gripper calculation
      temp_pos = map(temp_time, 0, serial_update_speed, servo_from_pos[1], servo_to_pos[1]);
      
      //gripper 
      if(servo_gripper_active == 1) {
        servo2.writeMicroseconds(temp_pos);
      }
    }
  }


  if (millis() > analog_read_target) {
    analog_read_target = millis() + analog_read_delay;
    uint16_t temp_read;
    for (uint8_t m = 0; m < 8; m++) {
      //temp_read = analogRead(motor_pot[m]);

      //position from serial is injected here <-----------------------
      temp_read = serial_position[m];
      temp_read = map(temp_read, 0, 1023, 0, 8191); //remap to teensy resolution

      if (temp_read > 50 && temp_read <= 8100) { //only pass temp read if the value makes sense
        analog_read_value[m] = temp_read;
      }
      input_target_position[m] = analog_read_value[m];
      //Serial.print(analog_read_value[m]);
      //Serial.print('\t');
    }

    //Servo input injected here <-------------------------
    servo_pos1 = map(input_target_position[0], servo1_range[0], servo1_range[1], servo1_micros[0], servo1_micros[1]);
    servo_pos2 = map(input_target_position[1], servo2_range[0], servo2_range[1], servo2_micros[0], servo2_micros[1]);
    servo_pos1 = constrain(servo_pos1, servo1_micros[0], servo1_micros[1]);
    servo_pos2 = constrain(servo_pos2, servo2_micros[0], servo2_micros[1]);

    //servo smoother
    //free value in buffer
    servo_history[0][0] = servo_history[0][1];
    servo_history[0][1] = servo_history[0][2];
    servo_history[1][0] = servo_history[1][1];
    servo_history[1][1] = servo_history[1][2];

    //refill buffer
    servo_history[0][2] = servo_pos1;
    servo_history[1][2] = servo_pos2;

    //calculate average
    int32_t temp_calc;
    temp_calc = servo_history[0][0] + servo_history[0][1] + servo_history[0][2];
    temp_calc /= 3;
    servo_pos1 = temp_calc;
    temp_calc = servo_history[1][0] + servo_history[1][1] + servo_history[1][2];
    temp_calc /= 3;
    servo_pos2 = temp_calc;

    //set to and from pos, and start time
    servo_from_pos[0] = servo_to_pos[0];
    servo_from_pos[1] = servo_to_pos[1];
    servo_to_pos[0] = servo_pos1;
    servo_to_pos[1] = servo_pos2;
    servo_start_time = millis();

    /*if (servo_rotation_active == 1) {
      servo1.writeMicroseconds(servo_pos1);
      }
      if (servo_gripper_active == 1) {
      servo2.writeMicroseconds(servo_pos2);
      }//*/

    //Serial.println("");

    if (motion_activated == 2) { //if motion is activated, but not yet reset, do so now
      Serial.println("resetting motion parameters");

      //set the new position for all motors
      for (uint8_t m = 0; m < number_of_motors; m++) {
        motor_current_position[m] = map(input_target_position[m], 0, 8191, 0, motor_range[m]);
        Serial.print(" ");
        Serial.print(motor_current_position[m]);
        motor_current_velocity[m] = 1; //reset velocity
      }
      Serial.println("");


      for (uint8_t m = 0; m < number_of_motors; m++) {
        Serial.print(" ");
        Serial.print(motor_current_position[m]);
      }
      Serial.println("");

      motion_activated = 3;
    }

    if (motion_activated == 1) { //extra delay
      motion_activated = 2;
    }
  }

  if (millis() > pid_read_target) { //pid functions (only P really, updates velocity as well, so run often)
    pid_read_target = millis() + pid_read_delay;
    for (uint8_t m = 0; m < 8; m++) {

      //calculate target velocity

      //Serial.print(input_target_position[m]); Serial.print('\t');
      //Serial.print(motor_target_position[m]); Serial.print('\t');
      //Serial.print(motor_current_position[m]); Serial.print('\t');
      //Serial.print(motor_target_offset[m]); Serial.print('\t');
      float temp_calc = float(motor_target_offset[m]); //input offset (how far off the position is to target)
      temp_calc *= pid_kp; //add proportional component to it
      motor_target_velocity[m] = long(temp_calc);
      if (motor_target_velocity[m] >  motor_max_velocity[m]) {
        motor_target_velocity[m] = motor_max_velocity[m];
      }
      if (motor_target_velocity[m] <  motor_max_velocity[m] * -1) {
        motor_target_velocity[m] = motor_max_velocity[m] * -1;
      }
      //Serial.print(motor_target_velocity[m]); Serial.print('\t');

      if (motor_target_velocity[m] >= 0 && motor_target_velocity[m] < motor_min_velocity[m]) { //limit positive minimum velocity
        motor_target_velocity[m] = motor_min_velocity[m];
      }
      if (motor_target_velocity[m] < 0 && motor_target_velocity[m] > motor_min_velocity[m] * -1) { //limit negative minimum velocity
        motor_target_velocity[m] = motor_min_velocity[m] * -1;
      }

      //calculate current velocity
      if (motor_target_velocity[m] > motor_current_velocity[m]) {
        motor_current_velocity[m] += pid_max_delta;
      }
      else if (motor_target_velocity[m] < motor_current_velocity[m]) {
        motor_current_velocity[m] -= pid_max_delta;
      }
      if (motor_current_velocity[m] == 0) {
        motor_current_velocity[m] = 1; //protect against 0
      }

      if (motor_current_velocity[m] >= 0) {
        motor_direction[m] = 1;
      }
      else {
        motor_direction[m] = -1;
      }

      //Serial.print(motor_current_velocity[m]); Serial.print('\t');

      //calculate step delay
      step_speed[m] = 1000000 / motor_current_velocity[m];
      //Serial.print(step_speed[m]); Serial.println('\t');
    }
  }


  if (millis() > activation_target) { //check button
    activation_target = millis() + activation_delay;
    uint8_t temp_state = digitalRead(activation_switch); //get button state

    //if falling edge
    if (activation_switch_history == 1 && temp_state == 0) {
      if (motion_activated == 0) {
        motion_activated = 1; //set motion to active
        Serial.println("motion enabled");
        //set the new position for all motors
        for (uint8_t m = 0; m < number_of_motors; m++) {
          motor_current_position[m] = map(input_target_position[m], 0, 8192, 0, motor_range[m]);
          motor_current_velocity[m] = 1; //reset velocity
        }
      }
      else {
        motion_activated = 0; //set motion to inactive
        Serial.println("motion disabled");
      }
    }

    //reset history
    activation_switch_history = temp_state;
  }
}

void SerialExecute() { //on 1 in serial update, get values and execute commands
  //inkjet_line_number = Ser.GetLineNumber(); //get line number
  serial_command = Ser.GetCommand(); //get command
  for (uint8_t s = 0; s < 8; s++) {
    serial_small_value[s] = Ser.GetSmallValue(s);  //get small value
  }

  if (serialCommandEcho == 1) {
    Serial.print("#COM:"); Serial.println(serial_command);
  }

  //check line number (lol, nope, not right now, why did I bother then with a linenumber? dunno, might actually disable it)
  //Later edit, I already disabled it.

  //Serial.print("Executing command: "); Serial.println(inkjet_command);
  switch (serial_command) { //check command, execute based on command
    case 1398035028: { //STRT, Start
        motion_activated = 1; //enable motion
        Serial.println("starting motion");
      } break;
    case 1398034256: { //STOP, Stop
        motion_activated = 0; //disable motion
        Serial.println("Stopping motion");
      } break;
    case 1095782227: { //APOS, Get position
        Serial.print("Getting analog positions:");
        for (uint8_t s = 0; s < 8; s++) {
          serial_position[s] = serial_small_value[s];
          Serial.print(" ");
          Serial.print(serial_position[s]);
        }
        Serial.println("");
      } break;
    case 1398035026: { //STRR, Start rotation
        servo_rotation_active = 1;
        Serial.println("Starting rotation");
      } break;
    case 1398034258: { //STOR, Stop rotation
        servo_rotation_active = 0;
        Serial.println("Stopping rotation");
      } break;
    case 1398035015: { //STRG, Start Gripper
        servo_gripper_active = 1;
        Serial.println("Starting gripper");
      } break;
    case 1398034247: { //STOG, Stop Gripper
        servo_gripper_active = 0;
        Serial.println("Stopping gripper");
      } break;




    default: {
        Serial.println("Unknown command");
      }
  }
}
