/*
   Serial block takes care of receiving and decoding the serial connection
   The serial port used is serial1

   Each block consists of a semi fixed structure, the default is:
   CCCC-SSSS-SSSS-SSSS-SSSS-SSSS-SSSS-SSSS-SSSSe

   The actual amount can change, but a full line may not be more than 64 bytes including end characters

   - is a whitespace, to indicate an end of a line
   C is a command. Each function has an up to 4 letter command
   S is a small number, in decimal
   e end character ('\r' of '\n')

  The list of commands is as follows:
  //direct inkjet commands
  -STRT: Start the motion
  -STOP: Stop the motion
  -APOS: Get analog position

*/

#include "Arduino.h"

#define max_read_length 130

class SerialCommand {
    //class member variables
    uint8_t serial_update_state = 0; //whether the outside needs to call functions
    uint8_t serial_first_line = 1; //whether the line received is the first line, used to reset line numbers
    int8_t error_flag = 0; //what error if any is happening
    uint8_t serial_external_state = 1; //set serial port to passthrough mode by default. (2 is input mode, but not yet implemented)
    uint8_t serial_external_passthrough_response = 0; //whether data received from the external serial needs to be passed to
#define external_serial_baudrate 9600
#define external_serial_update_delay 10 //in milliseconds
    uint32_t external_serial_update_target;

    //read variables
    char read_buffer[max_read_length]; //buffer for storing serial characters
    char decode_buffer[max_read_length]; //buffer for storing serial line
    uint16_t decode_buffer_pos = 0; //how much is in the read buffer
    uint16_t serial_available; //number of available bytes in hardware buffer

    int16_t end_character_pos; //where the end character is

    //write variables
    char write_buffer[64]; //buffer for handling serial output
    uint8_t write_characters; //how many characters are in the buffer to be sent out

    //latest decoded line values
    uint16_t serial_line_number;
    uint32_t serial_command;
    int32_t serial_small_value[8];
    uint8_t serial_raw[50];

    //debug
    uint8_t serial_debug_enabled;

  public:
    SerialCommand() {
      Serial1.begin(9600);
      write_characters = 0;
    }

    void Begin() { //sends start message to indicate live connection
      serial_debug_enabled = 0; //disable debug
    }

    int16_t Update() { //updates all serial command functions
      /*
         Read what is currently in the buffer
         if a full line is in buffer, read and decode it
         If the program needs to act on a new line, update will return a 1, else a 0
         returns a -negative on error
      */

      serial_available = Serial1.available(); //get number available
      if (serial_available) { //if there are lines in hardware buffer
        if (serial_debug_enabled == 1) {
          Serial.print("Available: ");
          Serial.println(serial_available);
        }

        serial_available = constrain(serial_available, 0, 64); //constrain lines available to hardware block size
        Serial1.readBytes(read_buffer, serial_available); //read bytes from hardware buffer
        //send received response
        //Serial.println("OK"); //print ok
        //Serial.send_now(); //send data ASAP

        //calculate decode buffer space left
        int16_t temp_space_left = max_read_length;
        temp_space_left -= decode_buffer_pos;
        temp_space_left -= serial_available;

        if (serial_debug_enabled == 1) {
          Serial.print("received: "); Serial.write(read_buffer, serial_available); Serial.println("");
          Serial.print("Spaceleft: "); Serial.println(temp_space_left);
        }

        uint16_t temp_read = 0;
        if (temp_space_left > 0) { //add read_buffer to serial_buffer if possible
          for (uint16_t b = decode_buffer_pos; b < decode_buffer_pos + serial_available; b++) { //add to read buffer
            decode_buffer[b] = read_buffer[temp_read];
            temp_read ++; //add one to readbuffer position
          }
          decode_buffer_pos += serial_available;
          if (serial_debug_enabled == 1) {
            Serial.print("Buffer: '"); for (uint16_t r = 0; r < max_read_length; r++) {
              Serial.print(decode_buffer[r]);
            } Serial.println("'");
            Serial.print("Decode buffer pos: "); Serial.println(decode_buffer_pos);
          }

          end_character_pos = -1; //reset end character pos
          for (uint16_t b = 0; b < max_read_length; b++) { //loop through buffer to find end character (start at front)
            if (IsEndCharacter(decode_buffer[b]) == 2) { //if end of line character
              end_character_pos = b; //set end character search
              break; //break from search loop
            }
          }

          uint16_t temp_start = 0, temp_end = 0; //keeps track of the blocks to decode
          uint8_t temp_character = 0; //keeps track of which character is being read
          uint8_t temp_read_error = 0, temp_ended = 0;
          uint32_t temp_decode;
          if (end_character_pos != -1) { //if end character found, start decoding from front
            if (serial_debug_enabled == 1) {
              Serial.print("End of line character found: ");
              Serial.println(end_character_pos);
            }

            //reset everything to 0 when reading starts
            //serial_line_number = 0;
            serial_command = 0;
            for (uint8_t s = 0; s < 8; s++) {
              serial_small_value[s] = 0;
            }

            //look for first break, command
            if (temp_read_error == 0 && temp_ended == 0) { //check if no error and not ended
              if (serial_debug_enabled == 1) {
                Serial.println("Start looking for command");
              }
              for (uint16_t e = temp_start; e <= end_character_pos; e++) { //look for second break, command
                if (IsEndCharacter(decode_buffer[e])) {
                  if (IsEndCharacter(decode_buffer[e]) == 2) {
                    temp_ended = 1;  //set ended to 1 if this was a line end, not a space
                    //if (serial_debug_enabled == 1){ Serial.println("Ended");}
                  }
                  if (serial_debug_enabled == 1) {
                    Serial.print("End char in CM: ");
                    Serial.println(e);
                  }
                  serial_command = 0; //reset command number
                  temp_end = e; //note the position where the line ends
                  for (int16_t d = temp_end - 1; d >= temp_start; d--) { //decode from end (LSB) to start (MSB)
                    temp_decode = decode_buffer[d];//find command (raw ascii to value)
                    temp_decode = temp_decode << 8 * temp_character; //store temp command
                    temp_character++;
                    serial_command |= temp_decode;
                  }
                  if (serial_debug_enabled == 1) {
                    Serial.print("Decoded CM: ");
                    Serial.println(serial_command);
                  }
                  temp_start = e + 1; //set new start
                  temp_character = 0;
                  break; //break from current for loop
                }
              }
            }

            //make while block for several small values ---------------------------------
            for (uint8_t s = 0; s < 8; s++) {
              //look for small value breaks,
              if (temp_ended == 1){ //if search is ended, break from loop
                break;
              }
              int8_t temp_signed = 1; //set sign to positive
              if (temp_read_error == 0 && temp_ended == 0) { //check if no error and not ended
                if (serial_debug_enabled == 1) Serial.println("Start looking for small value");
                for (uint16_t e = temp_start; e <= end_character_pos; e++) { //look for third break, small
                  if (IsEndCharacter(decode_buffer[e])) {
                    if (IsEndCharacter(decode_buffer[e]) == 2) {
                      temp_ended = 1;  //set ended to 1 if this was a line end, not a space
                      //if (serial_debug_enabled == 1){ Serial.println("Ended");}
                    }
                    if (serial_debug_enabled == 1) {
                      Serial.print("End char in SV: ");
                      Serial.println(e);
                    }
                    //look for first character being a sign
                    if (decode_buffer[temp_start] == '-') { //if first character is sign
                      if (serial_debug_enabled == 1) {
                        Serial.println("Sign found");
                      }
                      temp_start ++; //ignore first character
                      temp_signed = -1; //set sign to -1
                    }
                    serial_small_value[s] = 0; //reset small value number
                    temp_end = e; //note the position where the line ends
                    for (int16_t d = temp_end - 1; d >= temp_start; d--) { //decode from end (LSB) to start (MSB)
                      if (IsB64(decode_buffer[d])) { //if B64 value
                        temp_decode = B10Lookup(decode_buffer[d]); //decode lead number
                        for (uint8_t power = 0; power < temp_character; power++){ //add power to character
                          temp_decode *= 10;
                        }
                        temp_character++;
                        serial_small_value[s] += temp_decode; //add value to lead number
                      }
                      else { //if any non B64 char found, go to error
                        temp_read_error = 1;
                        if (serial_debug_enabled == 1) {
                          Serial.print("Error");
                        }
                      }
                    }
                    serial_small_value[s] = serial_small_value[s] * temp_signed; //add sign to value
                    if (serial_debug_enabled == 1) {
                      Serial.print("Decoded SV: ");
                      Serial.println(serial_small_value[s]);
                    }
                    temp_start = e + 1; //set new start
                    temp_character = 0;
                    break; //break from current for loop
                  }
                }
              }
            }

            //!!!WARNING, buffer removing untested----------------------------------------------------
            //remove read bytes from decode buffer
            if (serial_debug_enabled == 1) {
              Serial.println("Removing read lines");
            }
            int16_t temp_reset = end_character_pos;
            for (uint16_t r = 0; r < max_read_length; r++) {
              if (decode_buffer_pos > 0) {
                decode_buffer[r] = decode_buffer[temp_reset];
                decode_buffer_pos--; //remove one pos from decode buffer
                temp_reset++; //remove one to read value
              }
              else {
                decode_buffer[r] = 0;
              }
            }
            //Serial.print("Decode buffer pos: "); Serial.println(decode_buffer_pos);

            if (temp_read_error == 0 && temp_ended == 1) { //if ended and no mistakes
              return 1;  //return a 1
            }
            else {
              return 0; //mistake somewhere along the way
            }
          } //end of end character found if
        } //end of spaceleft > 0
        else { //if there was no space left in buffer
          return -1;
        }
      } //end of serial available
      return 0; //no new data, return a 0
    }
    uint16_t GetBufferLeft() { //gets how many characters are left in the buffer
      uint16_t temp_return = max_read_length - 1;
      temp_return -= decode_buffer_pos;
      return temp_return;
    }
    uint16_t GetLineNumber() { //returns the last read line number
      return serial_line_number;
    }
    uint32_t GetCommand() { //returns the last read command
      return serial_command;
    }
    int32_t GetSmallValue(uint8_t temp_val) { //returns the last read small value
      temp_val = constrain(temp_val, 0, 7);
      return serial_small_value[temp_val];
    }
    void GetRaw(uint8_t* temp_raw) { //takes an uint8_t[50] array and fills it with the last read value
      for (uint8_t r = 0; r < 50; r++) {
        temp_raw[r] = serial_raw[r];
      }
    }
    uint8_t IsEndCharacter(char temp_input) { //checks if a char is an end character, 1 for block end, 2 for line end
      if (temp_input == 10) return 2; //new line
      if (temp_input == 13) return 2; //carriage return
      if (temp_input == 32) return 1; //space
      return 0;
    }
    uint8_t IsB10(char temp_input) {  //reads a character and determines if it is a valid base 10 character
      if (temp_input >= '0' && temp_input <= '9') return 1;
      if (temp_input == '-') return 1; //negative is accepted in base 10 mode
      return 0; //no match found
    }
    uint8_t IsB64(char temp_input) { //reads a character and determines if it is a valid base 64 character
      if (temp_input >= 'A' && temp_input <= 'Z') return 1;
      if (temp_input >= 'a' && temp_input <= 'z') return 1;
      if (temp_input >= '0' && temp_input <= '9') return 1;
      if (temp_input == '+') return 1;
      if (temp_input == '/') return 1;
      return 0; //no match found
    }
    int8_t B10Lookup(char c) { //returns the 0-9 from ascii
      if (c >= '0' && c <= '9') return c - 48; //0-9
      return -1; //-1
    }
    int8_t B64Lookup(char c) { //Convert a base 64 character to decimal, (returns -1 if no valid character is found)
      if (c >= 'A' && c <= 'Z') return c - 'A'; //0-25
      if (c >= 'a' && c <= 'z') return c - 71; //26-51
      if (c >= '0' && c <= '9') return c + 4; //52-61
      if (c == '+') return 62; //62
      if (c == '/') return 63; //63
      return -1; //-1
    }
    char ToB64Lookup(uint8_t temp_input) {
      if (temp_input >= 0 && temp_input <= 25) return temp_input + 'A'; //0-25
      if (temp_input >= 26 && temp_input <= 51) return temp_input + 71; //26-51
      if (temp_input >= 52 && temp_input <= 61) return temp_input - 4; //52-61
      if (temp_input == 62) return '+';
      if (temp_input == 63) return '/';
      return '&'; //return '&' if false, this will trigger code reading it to mark as mistake
    }
    void WriteValueToB64(int32_t temp_input) { //adds a value to the writebuffer in B64
      //Serial.println("Encoding to B64");
      //Serial.print("Number: "); Serial.println(temp_input);
      uint8_t temp_sign = 0;
      uint8_t temp_6bit;
      char temp_reverse_array[5];
      if (temp_input < 0) { //if value is negative
        temp_sign = 1; //set negative flag to 1
        temp_input = abs(temp_input); //make the input positive
      }
      for (int8_t a = 0; a < 5; a++) { //loop through the input to convert it
        //Serial.print("Encoding: "); Serial.println(a);
        temp_6bit = 0; //reset 6 bit value
        temp_6bit = temp_input & 63; //read first 6 bits
        //Serial.print("Turns to: "); Serial.println(temp_6bit);
        temp_reverse_array[a] = ToB64Lookup(temp_6bit);
        temp_input = temp_input >> 6; //shift bits
      }
      //add reverse array to output buffer
      if (temp_sign == 1) { //add negative to value
        write_buffer[write_characters] = '-';
        write_characters++;
      }
      uint8_t temp_started = 0;
      for (int8_t a = 4; a >= 0; a--) {
        if (temp_started == 0) { //until a value is seen all zero's ('A''s) are ignored
          if (temp_reverse_array[a] != 'A') temp_started = 1; //start when first non 0 is found
          if (a == 0) temp_started = 1; //also start when the first character is reached
        }
        if (temp_started == 1) { //when started, write to output buffer
          write_buffer[write_characters] = temp_reverse_array[a];
          write_characters++;
        }
      }
    }
    void WriteTestArrayToB64(uint8_t temp_input[]) { //adds a test array to the writebuffer
      uint16_t temp_n = 0; //nozzle counter
      uint8_t temp_6bit; //6 bit decode value
      char temp_output[50]; //50 char output array
      for (uint8_t r = 0; r < 50; r++) {
        temp_output[r] = ' '; //reset array values
      }
      //decode 300 array to B64
      for (uint8_t B = 0; B < 50; B++) {
        temp_6bit = 0; //reset 6 bit value
        for (uint8_t b = 0; b < 6; b++) {
          //Serial.print("Decoding: "); Serial.print(temp_n); Serial.print(", Value: "); Serial.println(temp_input[temp_n]);
          temp_6bit |= (temp_input[temp_n] & 1) << b;
          temp_n++;
        }
        temp_output[B] = ToB64Lookup(temp_6bit); //write 6 bit value to output
        //Serial.print("Encoding: "); Serial.print(temp_6bit); Serial.print(", Value: "); Serial.println(temp_output[B]);
      }

      //write B64 to output buffer
      for (int8_t B = 49; B >= 0; B--) {
        write_buffer[write_characters] = temp_output[B];
        write_characters++;
      }
    }
    void DebugSet(uint8_t temp_input) {
      temp_input = constrain(temp_input, 0, 1);
      serial_debug_enabled = temp_input;
    }
    void SendResponse() { //will write the line in the buffer and mark it to be sent ASAP
      write_buffer[write_characters] = '\n'; //add carriage return
      write_characters++;
      Serial.write(write_buffer, write_characters);
      Serial.send_now(); //send all in the buffer ASAP
      write_characters = 0;//reset write counter
    }
    void PrintHelp() {
      Serial.print(
        "\n"
        "A command looks as follows: CCCC-PPPPP-RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRe\n"
        "- is a whitespace, to indicate an end of a line\n"
        "C = is a command. Each function has an up to 4 letter command\n"
        "P = is a small number in base 64, optional, can have a sign ('-')\n"
        "R = Raw data in base 64, optional, in inkjet reserved for inkjet data\n"
        "e = end character (carriage return or new line)\n"
        "\n"
        "Based on context, some blocks can be different, but by default all number carrying blocks will be encoded in base 64\n"
        "From 0 to 63: ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/\n"
        "Here, A is 0, and B is 1!\n"
        "\n"
        "List of commands:\n"
        "SBR: Send inkjet to buffer raw (needs small for pos and raw for inkjet data)\n"
        "SBT: Send inkjet to buffer toggle format (needs small for pos and raw for inkjet data)\n"
        "\n"
        "PHT: Preheat printhead (needs small for n pulses)\n"
        "PRM: Prime printhead (needs small for n pulses)\n"
        "THD: Test printhead (no extra input)\n"
        "GTP: Get temperature (no extra input)\n"
        "\n"
        "SPME: Set position mode to encoder (no extra input)\n"
        "SPMV: Set position mode to virtual (no extra input)\n"
        "\n"
        "SEP: Set encoder position (needs small for n position in microns)\n"
        "GEP: Get encoder position (no extra input)\n"
        "SDP: Set DPI (needs small for n DPI)\n"
        "SDN: Set density (needs small for n percentage)\n"
        "\n"
        "VENA: Virtual enable (needs small for state (1 or 0))\n"
        "SVV: Set virtual velocity (needs small for n velocity in mm/s)\n"
        "SVR: Set virtual position reset (needs small for n position in microns)\n"
        "VTRI: Virtual Trigger (no extra input)\n"
        "VSTO: Virtual stop (no extra input)\n"
        "STMA: Set trigger mode A (needs small for state (0 off, 1 rising, 2 falling, 3 toggle))\n"
        "STMB: Set trigger mode B (needs small for state (0 off, 1 rising, 2 falling, 3 toggle))\n"
        "STPU: Set trigger push notification (needs small for state (1 or 0))\n"
        "\n"
        "BRL: Buffer get read left (no extra input)\n"
        "BWL: Buffer get write left (no extra input)\n"
        "\n"
        "#COM: Command echo (needs small for state (1 or 0))\n"
        "#NDS: Numberdecode small (needs small to decode)\n"
        "#NDR: Number decode raw (needs small for mode (0 for binary, 1 for numbers) and raw to decode)\n"
        "#NES: Number encode small (needs small in decimal integer, encodes it to a base64 number)\n"
        "#DEB: Debug mode (needs small for debug mode)\n"
        "\n"
        "debug modes:\n"
        "0: all off\n"
        "1: all on\n"
        "2: Serial only\n"
        "\n"
      );
    }
    void RespondTestPrinthead() {

    }
    void RespondTemperature(int32_t temp_temp) {
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'G';
      write_buffer[1] = 'T';
      write_buffer[2] = 'P';
      write_buffer[3] = ':';
      WriteValueToB64(temp_temp); //convert temperature to 64 bit
      SendResponse(); //send temperature
    }
    void RespondEncoderPos(int32_t temp_pos) {
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'G';
      write_buffer[1] = 'E';
      write_buffer[2] = 'P';
      write_buffer[3] = ':';
      WriteValueToB64(temp_pos); //convert temperature to 64 bit
      SendResponse(); //send temperature
    }
    void RespondEncoderResolution(float temp_resolution) {
      int32_t temp_response = long(temp_resolution);
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'G';
      write_buffer[1] = 'E';
      write_buffer[2] = 'R';
      write_buffer[3] = ':';
      WriteValueToB64(temp_response); //convert temperature to 64 bit
      SendResponse(); //send left
    }
    void RespondBufferReadLeft(int32_t temp_left) {
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'B';
      write_buffer[1] = 'R';
      write_buffer[2] = 'L';
      write_buffer[3] = ':';
      WriteValueToB64(temp_left); //convert temperature to 64 bit
      SendResponse(); //send left
    }
    void RespondBufferWriteLeft(int32_t temp_left) {
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'B';
      write_buffer[1] = 'W';
      write_buffer[2] = 'L';
      write_buffer[3] = ':';
      WriteValueToB64(temp_left); //convert temperature to 64 bit
      SendResponse(); //send left
    }
    void RespondTestResults(uint8_t temp_mode, uint8_t temp_result[]) {
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'T';
      write_buffer[1] = 'H';
      write_buffer[2] = 'D';
      write_buffer[3] = ':';
      WriteTestArrayToB64(temp_result); //convert 1B array to 64 bit
      SendResponse(); //send left
    }
    void RespondVirtualPosition(int32_t temp_pos) {
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'G';
      write_buffer[1] = 'V';
      write_buffer[2] = 'P';
      write_buffer[3] = ':';
      WriteValueToB64(temp_pos); //convert 1B array to 64 bit
      SendResponse(); //send left
    }
    void RespondVirtualVelocity(int32_t temp_velocity) {
      write_characters = 4; //set characters to value after adding response header
      write_buffer[0] = 'G';
      write_buffer[1] = 'V';
      write_buffer[2] = 'V';
      write_buffer[3] = ':';
      WriteValueToB64(temp_velocity); //convert 1B array to 64 bit
      SendResponse(); //send left
    }
    void RespondRaw(uint8_t temp_input[50], uint8_t temp_size, uint8_t temp_mode) {
      Serial.print("#NDR:");
      for (int16_t n = temp_size - 1; n >= 0; n--) {
        if (temp_mode == 0) { //binary mode
          for (uint8_t b = 0; b < 6; b++) {
            Serial.print(bitRead(temp_input[n], b));
          }
        }
        else { //number mode
          Serial.print(temp_input[n]); Serial.print(" ");
        }
      }
      Serial.println("");
    }
    void RespondEncodeSmall(int32_t temp_input) { //takes a numeric input and encodes it back to base 64. Used for manual typing help, can only do up to 5 character numbers
      uint8_t temp_encode;
      uint32_t temp_response_value = 0;
      uint32_t temp_decimal = 1;
      char response_array[5];
      for (uint8_t r  = 0; r < 5; r++) { //reset all the numbers in the array
        response_array[r] = 'A';
      }
      for (uint8_t d = 0; d < 25; d += 6) { //look at the value in 6 bit chunks
        temp_encode = (temp_input >> d) & 0B00111111;
        if (temp_encode >= 52 && temp_encode <= 62) { //only act if it is a number
          temp_encode -= 52; //change to integer numbers
        }
        else {
          temp_encode = 0; //change to 0
        }
        //Serial.print(temp_encode); Serial.print(", ");
        temp_response_value += temp_encode * temp_decimal; //add to counting value
        temp_decimal *= 10; //move one order of magnitude
      }
      //Serial.print(": "); Serial.println(temp_response_value);
      Serial.print("#NES: "); //Serial.print(temp_response_value); Serial.print(", Base64: ");

      //decode value to base 64 small value
      for (uint8_t d = 0; d < 31; d += 6) { //loop through the entire number
        temp_response_value = temp_response_value >> d; //go to next 6 bits
        uint8_t temp_base = temp_response_value & 63; //isolate 6 lsb
        response_array[d / 6] = ToB64Lookup(temp_base); //convert number to base64 number
        //Serial.print(temp_base); Serial.print(", ");  Serial.println(ToB64Lookup(temp_base));
      }
      //Serial.println(response_array);

      uint8_t temp_number_found = 0; //the number to indicate that something else than a 0 ('A') was found
      //reverse and decode
      for (uint8_t r = 0; r < 6; r++) {
        if (response_array[5 - r] != 'A') { //something else than 0 found (used to get rid of the useless zeroes, like 0025 becomes 25)
          temp_number_found = 1; //now real numbers
        }
        if (temp_number_found == 1) { //if real numbers are found
          Serial.print(response_array[5 - r]); //print the character
        }
      }
      Serial.println("");
    }
};
